/*
 * Copy me if you can.
 * by 20h
 */

#include <u.h>
#include <libc.h>
#include <thread.h>
#include <fcall.h>
#include <9p.h>
#include <mp.h>
#include <auth.h>
#include <libsec.h>

enum
{
	Blocksize = 512,

	Cread = 1,
	Cwrite,
};

uvlong size;
char *dest;
char *user;
uint tag;
int debug;
int diskfd;
ulong time0;
static AESstate aes;
UserPasswd *passwd;

void
usage(void)
{
	fprint(2, "usage: %s [-Dd] [-m mtpt] [-s srvname]"
			  " file\n", argv0);
	threadexits("usage");
}

enum
{
	Qroot,
	Qctl,
	Qdata,
};

#define PATH(type, n)		((type)|((n)<<8))
#define TYPE(path)			((int)(path) & 0xFF)
#define NUM(path)			((uint)(path)>>8)

/* Taken from Russ Cox's cryptfs. */
void
initkey(void)
{
	uchar dig[20], ivec[20];
	DigestState *s;

	sha1((uchar*)passwd->passwd, strlen(passwd->passwd), dig, nil);
	s = sha1((uchar*)passwd->passwd, strlen(passwd->passwd), nil, nil);
	sha1((uchar*)passwd->passwd, strlen(passwd->passwd), ivec, s);
	setupAESstate(&aes, dig, 16, ivec);
}

void
bufencrypt(void *a, vlong len)
{
	AESstate s;

	s = aes;
	aesCBCencrypt(a, len, &s);
}

void
bufdecrypt(void *a, vlong len)
{
	AESstate s;

	s = aes;
	aesCBCdecrypt(a, len, &s);
}
/* End stolen code. */

vlong
getblocksize(void)
{
	Dir *d;
	vlong size;

	d = dirfstat(diskfd);
	if(d == nil)
		return -1;
	size = d->length;
	free(d);
	
	return size;
}

long
decpread(int fd, Req *r)
{
	vlong i, b, v, e;
	uchar *buf;

	i = r->ifcall.offset % (vlong)Blocksize;
	b = r->ifcall.offset - i;
	v = (r->ifcall.count - i) % (vlong)Blocksize;
	e = b + (r->ifcall.count - i - v);
	e -= b;

	buf = emalloc9p((e < Blocksize) ? Blocksize : e);

	if(i != 0) {
		if(pread(fd, buf, Blocksize, b) < 0)
			sysfatal("pread: %r");
		bufdecrypt(buf, Blocksize);
		memmove(r->ofcall.data, buf + i, i);
	}

	if(e != 0) {
		if(pread(fd, buf, e, b + i) < 0)
			sysfatal("pread: %r");
		bufdecrypt(buf, e);
		memmove(r->ofcall.data + i, buf, e);
	}

	if(v != 0) {
		if(pread(fd, buf, Blocksize, b + i + e) < 0)
			sysfatal("pread: %r");
		bufdecrypt(buf, Blocksize);
		memmove(r->ofcall.data + i + e, buf, v);
	}

	r->ofcall.count = r->ifcall.count;
	free(buf);

	return 0;
}

long
encpwrite(int fd, Req *r)
{
	vlong i, b, v, e;
	uchar *buf;

	i = r->ifcall.offset % (vlong)Blocksize;
	b = r->ifcall.offset - i;
	v = (r->ifcall.count - i) % (vlong)Blocksize;
	e = b + (r->ifcall.count - i - v);
	e -= b;

	buf = emalloc9p((e < Blocksize) ? Blocksize : e);

	if(i != 0) {
		if(pread(fd, buf, Blocksize, b) < 0)
			sysfatal("pread: %r");
		memmove(buf + i, r->ifcall.data, i);
		bufencrypt(buf, Blocksize);
		if(pwrite(fd, buf, Blocksize, b) < 0)
			sysfatal("pwrite: %r");
	}

	if(e != 0) {
		bufencrypt(r->ifcall.data + i, e);
		if(pwrite(fd, r->ifcall.data + i, e, b + i) < 0)
			sysfatal("pwrite: %r");
	}

	if(v != 0) {
		if(pread(fd, buf, Blocksize, b + i + e) < 0)
			sysfatal("pread: %r");
		memmove(buf, r->ifcall.data + i + e, v);
		bufencrypt(buf, Blocksize);
		if(pwrite(fd, buf, Blocksize, b + i + e) < 0)
			sysfatal("pwrite: %r");
	}

	r->ofcall.count = r->ifcall.count;
	free(buf);

	return 0;
}

void
doblockreq(Req *r, int cmd)
{

	if(r->ifcall.offset > size) {
		respond(r, "block device is smaller");
		return;
	}

	if(r->ifcall.count + r->ifcall.offset >= size)
		r->ifcall.count = size - r->ifcall.offset;
	if(r->ifcall.count == 0) {
		respond(r, nil);
		return;
	}

	if(cmd == Cwrite)
		encpwrite(diskfd, r);
	else
		decpread(diskfd, r);

	respond(r, nil);
}

typedef struct Tab Tab;
struct Tab
{
	char *name;
	ulong mode;
};

Tab tab[] =
{
	".",		DMDIR|0555,
	"ctl",		0666,
	"data",		0777,
};

static void
fillstat(Dir *d, uvlong path)
{
	Tab *t;

	memset(d, 0, sizeof(*d));
	d->uid = estrdup9p(user);
	d->gid = estrdup9p(user);
	d->qid.path = path;
	d->atime = d->mtime = time0;
	t = &tab[TYPE(path)];
	d->name = estrdup9p(t->name);
	if(!strncmp(d->name, "data", 4))
		d->length = size;
	else
		d->length = 0;
	d->qid.type = t->mode>>24;
	d->mode = t->mode;
}

static void
fsattach(Req *r)
{
	if(r->ifcall.aname && r->ifcall.aname[0]){
		respond(r, "invalid attach specifier");
		return;
	}
	r->fid->qid.path = PATH(Qroot, 0);
	r->fid->qid.type = QTDIR;
	r->fid->qid.vers = 0;
	r->ofcall.qid = r->fid->qid;
	respond(r, nil);
}

static void
fsstat(Req *r)
{
	fillstat(&r->d, r->fid->qid.path);
	respond(r, nil);
}

static int
rootgen(int i, Dir *d, void*)
{
	i += Qroot + 1;

	if(i <= Qdata){
		fillstat(d, i);
		return 0;
	}
	return -1;
}

static char*
fswalk1(Fid *fid, char *name, Qid *qid)
{
	int i;
	ulong path;

	path = fid->qid.path;
	if(!(fid->qid.type & QTDIR))
		return "walk in non-directory";

	if(strcmp(name, "..") == 0){
		switch(TYPE(path)){
		case Qroot:
			return nil;
		default:
			return "bug in fswalk1";
		}
	}

	i = TYPE(path) + 1;
	while(i < nelem(tab)){
		if(strcmp(name, tab[i].name) == 0){
			qid->path = PATH(i, NUM(path));
			qid->type = tab[i].mode>>24;
			return nil;
		}
		if(tab[i].mode & DMDIR)
			break;
		i++;
	}
	return "directory entry not found";
}

static void
fsopen(Req *r)
{
	respond(r, nil);
}

static void
fsread(Req *r)
{
	char e[ERRMAX], buf[8192];
	ulong path;
	int n;

	path = r->fid->qid.path;
	switch(TYPE(path)){
	case Qroot:
		dirread9p(r, rootgen, nil);
		respond(r, nil);
		break;
	case Qctl:
		n = snprint(buf, sizeof(buf), "inquiry encrypted device\n"
					"geometry %lld %d\n", size, Blocksize);
		readbuf(r, buf, n);
		respond(r, nil);
		break;
	case Qdata:
		doblockreq(r, Cread);
		break;
	default:
		snprint(e, sizeof(e), "bug in fsread path=%lux", path);
		respond(r, e);
		break;
	}
}

static void
fswrite(Req *r)
{
	char e[ERRMAX];
	ulong path;

	path = r->fid->qid.path;
	switch(TYPE(path)){
	case Qctl:
		respond(r, nil);
		break;
	case Qdata:
		doblockreq(r, Cwrite);
		break;
	default:
		snprint(e, sizeof(e), "bug in fswrite path=%lux", path);
		respond(r, e);
		break;
	}
}

Srv fs = 
{
.attach=	fsattach,
.walk1=		fswalk1,
.open=		fsopen,
.read=		fsread,
.write=		fswrite,
.stat=		fsstat,
};

void
threadmain(int argc, char **argv)
{
	char *mtpt, *srvname;

	mtpt = "/n/cblock";
	srvname = nil;

	ARGBEGIN {
	case 'D':
		chatty9p++;
		break;
	case 'd':
		debug++;
		break;
	case 's':
		srvname = EARGF(usage());
		break;
	case 'm':
		mtpt = EARGF(usage());
		break;
	default:
		usage();
	} ARGEND;

	if(argc < 1)
		usage();

	diskfd = open(argv[0], ORDWR);
	if(diskfd < 0)
		sysfatal("open: %r");

	passwd = auth_getuserpasswd(auth_getkey, 
		"proto=pass service=cblockfs file=%s", argv[0]);
	if(passwd == nil)
		sysfatal("auth_getuserpasswd: %r");
	initkey();

	time0 = time(0);
	user = getuser();
	size = getblocksize();

	threadpostmountsrv(&fs, srvname, mtpt, MBEFORE);
	threadexits(0);
}
